/**********************************************************

This software is part of J.-S. Caux's ABACUS library.

Copyright (c).

-----------------------------------------------------------

File:  JSC.h

Purpose:  Core header file, includes all descendents.


***********************************************************/

#ifndef _JSC_H_
#define _JSC_H_

// This core header file includes all the others

const char JSC_VERSION[20] = "ABACUS++G_8";

// Standard includes
#include <cmath>    
#include <complex>  // for complex number algebra
#include <string>

#include <iostream>
#include <fstream> 
#include <iomanip>

#include <cstdlib>  // for exit(), ...
#include <ctime>    // for clock(), ...
#include <limits>   // for numeric limits
#include <climits>   // for numeric limits

// For file sizes, etc
#include "sys/types.h"
#include "sys/stat.h"
#include "unistd.h"

// Signal handling
#include <stdio.h>     /* standard I/O functions                         */
#include <unistd.h>    /* standard unix functions, like getpid()         */
#include <sys/types.h> /* various type definitions, like pid_t           */
#include <signal.h>    /* signal name macros, and the signal() prototype */

// My own math functions and shortcuts
#include "JSC_util.h"

// Vectors and matrices
#include "JSC_Vect.h"   // My vector class definitions
#include "JSC_Matrix.h"  // My matrix class definitions

// Choose_Table
#include "JSC_Combi.h"

// Fitting, interpolating
#include "JSC_Fitting.h"

// Young tableaux
#include "JSC_Young.h" 

// Integration
#include "JSC_Integ.h"

// Special functions:
#include "JSC_Spec_Fns.h"

//***  Integrable models:

// Heisenberg spin-1/2 antiferromagnet
#include "JSC_Heis.h"      

// Lieb-Liniger
#include "JSC_LiebLin.h" 

// One-d spinless fermions:
#include "JSC_ODSLF.h"

// General:
//#include "JSC_Bethe.h"

// Thermodynamic Bethe Ansatz utilities
#include "JSC_TBA.h"

// State ensembles
#include "JSC_State_Ensemble.h"

// XXX in zero field:  Uq(sl(2)) stuff
#include "JSC_XXX_h0.h"

// XXZ in zero field:  quantum groups
#include "JSC_XXZ_h0.h"

// Two-component Bose gas
//#include "2CBG.h"

// Richardson 
//#include "Richardson.h"

// *** Correlation functions:

// New scanning protocols for ABACUS++
#include "JSC_Scan.h"

// Functions for everybody
//#include "JSC_fns.h"            // KEEP THIS INCLUDE LAST, SINCE IT USES PREVIOUS DECLARATIONS

// Numerical RG:
#include "JSC_NRG.h"

// OpenMP 
#include <omp.h>

// Typedefs:
typedef double DP;


#endif // _JSC_H_


