/**********************************************************

This software is part of J.-S. Caux's ABACUS library.

Copyright (c).

-----------------------------------------------------------

File:  LiebLin_Tgt0.cc

Purpose:  Finite temperature correlations for Lieb-Liniger


***********************************************************/

#include "JSC.h"

using namespace std;
using namespace JSC;

namespace JSC {


  /*
  DP Entropy (LiebLin_Bethe_State& RefState, DP epsilon)
  {
    // This function calculates the entropy of a finite Lieb-Liniger state,
    // using the quantum number space particle and hole densities.

    // The densities are a sum of Gaussians with width epsilon.

    // We calculate \rho (x) and \rho_h (x) for x_j on a fine lattice.

    int Ix2max = RefState.Ix2.max();
    int Ix2min = RefState.Ix2.min();

    //cout << Ix2max << "\t" << Ix2min << endl;

    // Give a bit of room for leftmost and rightmost particles Gaussians to attain asymptotic values:
    Ix2max += int(10.0 * RefState.L * epsilon);
    Ix2min -= int(10.0 * RefState.L * epsilon);

    //cout << Ix2max << "\t" << Ix2min << endl;

    DP xmax = Ix2max/(2.0* RefState.L);
    DP xmin = Ix2min/(2.0* RefState.L);

    //xmax += 10.0* epsilon;
    //xmin -= 10.0* epsilon;
    
    DP dx = 0.1/RefState.L;
    int Nptsx = int((xmax - xmin)/dx); 

    Vect<bool> occupied (false, (Ix2max - Ix2min)/2 + 1); // whether there is a particle or not
    for (int i = 0; i < RefState.N; ++i) occupied[(RefState.Ix2[i] - Ix2min)/2] = true;

    Vect<double> rho(0.0, Nptsx);
    Vect<double> rhoh(0.0, Nptsx);
    
    //cout << xmin << "\t" << xmax << "\t" << dx << "\t" << Nptsx << endl;

    DP x;
    DP epsilonsq = epsilon * epsilon;
    DP twoepsilonsq = 2.0 * epsilon * epsilon;

    //Vect<DP> xparticle(RefState.N);
    //for (int i = 0; i < RefState.N; ++i) xparticle[i] = RefState.Ix2[i]/(2.0* RefState.L);
    Vect<DP> xarray((Ix2max - Ix2min)/2 + 1);
    for (int i = 0; i < (Ix2max - Ix2min)/2 + 1; ++i) xarray[i] = (0.5*Ix2min + i)/RefState.L;

    for (int ix = 0; ix < Nptsx; ++ix) {
      x = xmin + dx * (ix + 0.5);
      //for (int i = 0; i < RefState.N; ++i) rho[ix] += 1.0/((x - xparticle[i]) * (x - xparticle[i]) + epsilonsq);
      //rho[ix] *= epsilon/(PI * RefState.L);
      for (int i = 0; i < (Ix2max - Ix2min)/2 + 1; ++i) {
	// Using Gaussians:
	//if (occupied[i]) rho[ix] += exp(-(x - xarray[i]) * (x - xarray[i])/twoepsilonsq);
	//else rhoh[ix] += exp(-(x - xarray[i]) * (x - xarray[i])/twoepsilonsq);
	// Using Lorentzians:
	if (occupied[i]) rho[ix] += 1.0/((x - xarray[i]) * (x - xarray[i]) + epsilonsq);
	else rhoh[ix] += 1.0/((x - xarray[i]) * (x - xarray[i]) + epsilonsq);
      }
      //rho[ix] /= sqrt(twoPI) * epsilon * RefState.L;
      //rhoh[ix] /= sqrt(twoPI) * epsilon * RefState.L;
      rho[ix] *= epsilon/(PI * RefState.L);
      rhoh[ix] *= epsilon/(PI * RefState.L);
      //cout << ix << " x = " << x << "\trho = " << rho[ix] << "\trhoh = " << rhoh[ix] << "\trho + rhoh = " << rho[ix] + rhoh[ix] << endl;
    }

    // Now calculate the entropy:
    complex<DP> entropy = 0.0;
    DP Deltax = log(RefState.L)/RefState.L;
    for (int ix = 0; ix < Nptsx; ++ix)
      //entropy -= ln_Gamma (1.0 + rho[ix]) + ln_Gamma (2.0 - rho[ix]); // This is ln (\rho_tot choose \rho) with \rho_tot = 1. 
      entropy += ln_Gamma (RefState.L * (rho[ix] + rhoh[ix]) * Deltax + 1.0) - ln_Gamma(RefState.L * rho[ix] * Deltax + 1.0) - ln_Gamma (RefState.L * rhoh[ix] * Deltax + 1.0); // This is ln (\rho_tot choose \rho) with \rho_tot = 1. 
    entropy *= dx/Deltax;

    //cout << "Entropy found " << entropy << "\t" << real(entropy) << endl;

    return(real(entropy));
  }
  */

  /*
  DP Entropy_rho (LiebLin_Bethe_State& RefState, int Delta)
  {
    // This function calculates the discrete entropy of a finite Lieb-Liniger state,
    // counting the possible permutations within windows of fixed width.

    // We assume that the quantum numbers are ordered.  

    // Calculate a density \rho_\Delta (x) for x_j on the lattice:

    int iK_UL = JSC::max(RefState.Ix2[0], RefState.Ix2[RefState.N - 1]);

    Vect<double> rhoD(0.0, 2* iK_UL);
    
    int fourDeltasq = 4 * Delta * Delta;
    for (int ix = 0; ix < 2* iK_UL; ++ix) {
      // x = (-iK_UL + 1/2 + ix)/L
      for (int i = 0; i < RefState.N; ++i) rhoD[ix] += 1.0/((-2*iK_UL + 1 + 2*ix -RefState.Ix2[i]) * (-2*iK_UL + 1 + 2*ix -RefState.Ix2[i]) + fourDeltasq); 
      rhoD[ix] *= 4.0 * Delta/PI;
      //cout << "x = " << (-iK_UL + 1/2 + ix)/RefState.L << "\trhoD = " << rhoD[ix] << endl;
    }

    // Now calculate the entropy:
    DP entropy = 0.0;
    for (int ix = 0; ix < 2* iK_UL; ++ix)
      entropy -= rhoD[ix] * log(rhoD[ix]) + (1.0 - rhoD[ix]) * log(1.0 - rhoD[ix]);

    return(entropy);
  }
  */

  DP Entropy_Fixed_Delta (LiebLin_Bethe_State& RefState, int Delta)
  {
    // This function calculates the discrete entropy of a finite Lieb-Liniger state,
    // counting the possible permutations within windows of fixed width.

    // We assume that the quantum numbers are ordered.  

    // Fill in vector of occupancies:
    int nrIs = (RefState.Ix2[RefState.N-1] - RefState.Ix2[0])/2 + 1 + 2*Delta; // assume Ix2 are ordered, leave space of Delta on both sides
    Vect<int> occupancy(0, nrIs);  // leave space of Delta on both sides
    for (int i = 0; i < RefState.N; ++i) occupancy[Delta + (RefState.Ix2[i] -RefState.Ix2[0])/2] = 1;
    // Check:
    int ncheck = 0;
    for (int i = 0; i < nrIs; ++i) if(occupancy[i] == 1) ncheck++;
    //cout << "Check occupancy: " << endl << occupancy << endl;
    if (ncheck != RefState.N) {
      cout << ncheck << "\t" << RefState.N << endl;
      JSCerror("Counting q numbers incorrectly in Entropy.");
    }

    // Define some useful numbers:
    Vect_DP oneoverDeltalnchoose(Delta + 1);
    for (int i = 0; i <= Delta; ++i) oneoverDeltalnchoose[i] = ln_choose(Delta, i)/Delta;
    //Vect_DP oneoverDeltalnchoosecheck(Delta + 1);
    //for (int i = 0; i <= Delta; ++i) oneoverDeltalnchoosecheck[i] = log(DP(choose(Delta, i)))/Delta;
    //cout << oneoverDeltalnchoose << endl;
    //cout << oneoverDeltalnchoosecheck << endl;

    // Compute entropy:
    DP entropy = 0.0;
    int np;
    for (int i = 0; i < nrIs - Delta; ++i) {
      np = 0;
      for (int iD = 0; iD < Delta; ++iD) if (occupancy[i + iD] == 1) np++;

      entropy += oneoverDeltalnchoose[np];
    }

    return(entropy);
  }
  
  DP Entropy (LiebLin_Bethe_State& RefState, int Delta)
  {
    // Perform an average of entropies for regulators from Delta to 2Delta:
    DP entropysum = 0.0;
    for (int ie = 0; ie < Delta; ++ie) entropysum += Entropy_Fixed_Delta (RefState, Delta + ie);
    return(entropysum/Delta);
  }

  DP Entropy (LiebLin_Bethe_State& RefState)
  {
    int Delta = int(log(RefState.L));
    return(Entropy (RefState, Delta));
  }
  /*
  DP Canonical_Free_Energy (LiebLin_Bethe_State& RefState, DP kBT, int Delta)
  {
    return(RefState.E - kBT * Entropy (RefState, Delta));
  }

  DP Canonical_Free_Energy (LiebLin_Bethe_State& RefState, DP kBT, int Delta)
  {
    return(RefState.E - kBT * Entropy (RefState, Delta));
  }
  */
  DP Canonical_Free_Energy (LiebLin_Bethe_State& RefState, DP kBT)
  {
    return(RefState.E - kBT * Entropy (RefState));
  }

  /*
  //LiebLin_Bethe_State Canonical_Saddle_Point_State (DP c_int, DP L, int N, DP kBT, int Delta)
  LiebLin_Bethe_State Canonical_Saddle_Point_State_pre20110618 (DP c_int, DP L, int N, DP kBT, DP epsilon)
  {
    // This function returns the discretized state minimizing the canonical free energy
    // F = E - T S.

    LiebLin_Bethe_State spstate(c_int, L, N);
    spstate.Compute_All (true);

    if (kBT < 1.0e-3) return(spstate); // interpret as zero T case

    LiebLin_Bethe_State spstateup = spstate;
    LiebLin_Bethe_State spstatedown = spstate;

    bool converged = false;
    bool convergedati = false;

    DP canfreeenstay, canfreeenup, canfreeendown;
    DP Estay, Eup, Edown;
    DP Sstay, Sup, Sdown;

    while (!converged) {

      spstate.Compute_All (false);
      converged = true;  // set to false if we change anything

      // If we can minimize the free energy by changing the quantum number, we do:
      // NOTE:  we keep the state symmetric w/r to parity.

      for (int i = 0; i < N/2; ++i) {
	// Try to increase or decrease this quantum number

	convergedati = false;

	while (!convergedati) {
	  
	  convergedati = true; // set to false if we change anything

	  Estay = spstate.E;
	  //Sstay = Entropy (spstate, Delta);
	  Sstay = Entropy (spstate, epsilon);
	  //canfreeenstay = Canonical_Free_Energy (spstate, kBT, Delta);
	  canfreeenstay = Estay - kBT * Sstay;

	  spstateup = spstate;
	  if (i == 0 || spstateup.Ix2[i-1] < spstateup.Ix2[i] - 2) {
	    spstateup.Ix2[i] -= 2;
	    spstateup.Ix2[N-1-i] += 2;
	    spstateup.Compute_All(false);
	    Eup = spstateup.E;
	    //Sup = Entropy(spstateup, Delta);
	    Sup = Entropy(spstateup, epsilon);
	    //canfreeenup = Canonical_Free_Energy (spstateup, kBT, Delta);
	    canfreeenup = Eup - kBT * Sup;
	  }
	  else canfreeenup = canfreeenstay + 1.0e-6;
	  
	  spstatedown = spstate;
	  if (spstatedown.Ix2[i+1] > spstatedown.Ix2[i] + 2) {
	    spstatedown.Ix2[i] += 2;
	    spstatedown.Ix2[N-1-i] -= 2;
	    spstatedown.Compute_All(false);
	    Edown = spstatedown.E;
	    //Sdown = Entropy(spstatedown, Delta);
	    Sdown = Entropy(spstatedown, epsilon);
	    //canfreeendown = Canonical_Free_Energy (spstatedown, kBT, Delta);
	    canfreeendown = Edown - kBT * Sdown;
	  }
	  else canfreeendown = canfreeenstay + 1.0e-6;
	  
	  //cout << "i = " << i << "\t" << spstate.Ix2[i] << "\t\t" << canfreeenstay << "\t" << canfreeenup << "\t" << canfreeendown 
	  // << "\t\t" << Estay << "\t" << Eup << "\t" << Edown << "\t\t" << Sstay << "\t" << Sup << "\t" << Sdown << endl;
	  // Choose what to do:
	  if (canfreeenup < canfreeenstay && canfreeendown < canfreeenstay) 
	    cout << canfreeenstay << "\t" << canfreeenup << "\t" << canfreeendown << "\tWarning:  unclear option for minimization." << endl;
	  else if (canfreeenup < canfreeenstay) {
	    spstate = spstateup;
	    convergedati = false;
	    converged = false;
	  }
	  else if (canfreeendown < canfreeenstay) {
	    spstate = spstatedown;
	    convergedati = false;
	    converged = false;
	  }
	  // else do nothing.
	} // while !convergedati

      } // for i
    } // while (!converged)

    return(spstate);
  }
  */

  /* REMOVED FROM ++G_3 ONWARDS
  //LiebLin_Bethe_State Canonical_Saddle_Point_State (DP c_int, DP L, int N, DP kBT, int Delta)
  //LiebLin_Bethe_State Canonical_Saddle_Point_State (DP c_int, DP L, int N, DP kBT, DP epsilon)
  LiebLin_Bethe_State Canonical_Saddle_Point_State_pre8 (DP c_int, DP L, int N, DP kBT)
  {
    // This function returns the discretized state minimizing the canonical free energy
    // F = E - T S.

    // Improvement on version pre20110618: allow for pairwise `in' and `out' movements
    // keeping the energy the same to order 1/L^2 but changing the entropy.

    // The regulator Delta also has become an `internal' parameter.

    LiebLin_Bethe_State spstate(c_int, L, N);
    spstate.Compute_All (true);

    if (kBT < 1.0e-3) return(spstate); // interpret as zero T case

    LiebLin_Bethe_State spstateup = spstate;
    LiebLin_Bethe_State spstatedown = spstate;
    LiebLin_Bethe_State spstatein = spstate;
    LiebLin_Bethe_State spstateout = spstate;

    bool converged = false;
    bool convergedati = false;

    DP canfreeenstay, canfreeenup, canfreeendown, canfreeenin, canfreeenout;
    DP Estay, Eup, Edown, Ein, Eout;
    DP Sstay, Sup, Sdown, Sin, Sout;
    Estay = 0.0; Eup = 0.0; Edown = 0.0; Ein = 0.0; Eout = 0.0;
    Sstay = 0.0; Sup = 0.0; Sdown = 0.0; Sin = 0.0; Sout = 0.0;

    while (!converged) {

      spstate.Compute_All (false);
      converged = true;  // set to false if we change anything

      // If we can minimize the free energy by changing the quantum number, we do:
      // NOTE:  we keep the state symmetric w/r to parity.

      for (int i = 0; i < N/2 - 1; ++i) {

	// Try to increase or decrease the quantum numbers at i , 
	// or do an (approximately) energy-preserving move with i+1 (and parity pairs)
	// giving 5 possible options: stay same, (\pm 1, 0), (+1, -1), (-1, +1)

	convergedati = false;

	while (!convergedati) {
	  
	  convergedati = true; // set to false if we change anything

	  Estay = spstate.E;
	  Sstay = Entropy (spstate);
	  //Sstay = Entropy (spstate, Delta);
	  //Sstay = Entropy (spstate, epsilon);
	  //canfreeenstay = Canonical_Free_Energy (spstate, kBT, Delta);
	  canfreeenstay = Estay - kBT * Sstay;

	  spstateup = spstate;
	  if (i == 0 || spstateup.Ix2[i-1] < spstateup.Ix2[i] - 2) {
	    spstateup.Ix2[i] -= 2;
	    spstateup.Ix2[N-1-i] += 2;
	    spstateup.Compute_All(false);
	    Eup = spstateup.E;
	    Sup = Entropy(spstateup);
	    //Sup = Entropy(spstateup, Delta);
	    //Sup = Entropy(spstateup, epsilon);
	    //canfreeenup = Canonical_Free_Energy (spstateup, kBT, Delta);
	    canfreeenup = Eup - kBT * Sup;
	  }
	  else canfreeenup = canfreeenstay + 1.0e-6;
	  
	  spstatedown = spstate;
	  if (spstatedown.Ix2[i+1] > spstatedown.Ix2[i] + 2) {
	    spstatedown.Ix2[i] += 2;
	    spstatedown.Ix2[N-1-i] -= 2;
	    spstatedown.Compute_All(false);
	    Edown = spstatedown.E;
	    Sdown = Entropy(spstatedown);
	    //Sdown = Entropy(spstatedown, Delta);
	    //Sdown = Entropy(spstatedown, epsilon);
	    //canfreeendown = Canonical_Free_Energy (spstatedown, kBT, Delta);
	    canfreeendown = Edown - kBT * Sdown;
	  }
	  else canfreeendown = canfreeenstay + 1.0e-6;

	  spstatein = spstate;
	  if (spstatein.Ix2[i+1] > spstatein.Ix2[i] + 4) { // can move them closer
	    spstatein.Ix2[i] += 2;
	    spstatein.Ix2[i+1] -= 2;
	    spstatein.Ix2[N-1-i] -= 2;
	    spstatein.Ix2[N-1-i-1] += 2;
	    spstatein.Compute_All(false);
	    Ein = spstatein.E;
	    Sin = Entropy(spstatein);
	    //Sin = Entropy(spstatein, Delta);
	    //Sin = Entropy(spstatein, epsilon);
	    //canfreeenin = Canonical_Free_Energy (spstatein, kBT, Delta);
	    canfreeenin = Ein - kBT * Sin;
	  }
	  else canfreeenin = canfreeenstay + 1.0e-6;

	  spstateout = spstate;
	  if (i == 0 && spstateout.Ix2[1] + 2 < spstateout.Ix2[2] 
	      || (i < N/2 - 1 && spstateout.Ix2[i] - 2 > spstateout.Ix2[i-1] 
		  && spstateout.Ix2[i+1] + 2 < spstateout.Ix2[i+2])) { // can move them further apart
	    spstateout.Ix2[i] -= 2;
	    spstateout.Ix2[i+1] += 2;
	    spstateout.Ix2[N-1-i] += 2;
	    spstateout.Ix2[N-1-i-1] -= 2;
	    spstateout.Compute_All(false);
	    Eout = spstateout.E;
	    Sout = Entropy(spstateout);
	    //Sout = Entropy(spstateout, Delta);
	    //Sout = Entropy(spstateout, epsilon);
	    //canfreeenout = Canonical_Free_Energy (spstateout, kBT, Delta);
	    canfreeenout = Eout - kBT * Sout;
	  }
	  else canfreeenout = canfreeenstay + 1.0e-6;
	  
	  //cout << setprecision(8) << "i = " << i << "\t" << spstate.Ix2[i] << "\t" << spstate.Ix2[i+1] << "\t\t" << canfreeenstay << "\t" << canfreeenup << "\t" << canfreeendown << "\t" << canfreeenin << "\t" << canfreeenout << endl;
	  //cout << "\t\tE: " << Estay << "\t" << Eup << "\t" << Edown << "\t" << Ein << "\t" << Eout << endl;
	  //cout << "\t\tS: " << Sstay << "\t" << Sup << "\t" << Sdown << "\t" << Sin << "\t" << Sout << endl;

	  // Choose what to do: find minimum, 
	  if (canfreeenstay < canfreeenup && canfreeenstay < canfreeendown && canfreeenstay < canfreeenin && canfreeenstay < canfreeenout) {
	    // do nothing, convergetati is already true
	  }
	  else if (canfreeenup < canfreeenstay && canfreeenup < canfreeendown && canfreeenup < canfreeenin && canfreeenup < canfreeenout) {
	    spstate = spstateup;
	    convergedati = false;
	    converged = false;
	  }
	  else if (canfreeendown < canfreeenstay && canfreeendown < canfreeenup && canfreeendown < canfreeenin && canfreeendown < canfreeenout) {
	    spstate = spstatedown;
	    convergedati = false;
	    converged = false;
	  }
	  else if (canfreeenin < canfreeenstay && canfreeenin < canfreeenup && canfreeenin < canfreeendown && canfreeenin < canfreeenout) {
	    spstate = spstatein;
	    convergedati = false;
	    converged = false;
	  }
	  else if (canfreeenout < canfreeenstay && canfreeenout < canfreeenup && canfreeenout < canfreeendown && canfreeenout < canfreeenin) {
	    spstate = spstateout;
	    convergedati = false;
	    converged = false;
	  }
	  else cout << canfreeenstay << "\t" << canfreeenup << "\t" << canfreeendown << "\t" << canfreeenin << "\t" << canfreeenout << "\tWarning:  unclear option for minimization." << endl;

	  // else do nothing.
	} // while !convergedati

      } // for i
    } // while (!converged)

    //cout << "Number of holes between I's: " << endl;
    //for (int i = 0; i < N/2; ++i) cout << (spstate.Ix2[i+1] - spstate.Ix2[i])/2 - 1 << "\t";
    //cout << endl;

    return(spstate);
  }
  */

  DP rho_of_lambdaoc_1 (LiebLin_Bethe_State& RefState, DP lambdaoc, DP delta)
  {
    DP answer = 0.0;
    for (int i = 0; i < RefState.N; ++i)
      answer += atan((lambdaoc - RefState.lambdaoc[i])/delta + 0.5) - atan((lambdaoc - RefState.lambdaoc[i])/delta - 0.5);
    answer *= 1.0/(PI * delta * RefState.L);

    return(answer);
  }

  DP rho_of_lambdaoc_2 (LiebLin_Bethe_State& RefState, DP lambdaoc, DP delta)
  {
    DP answer = 0.0;
    for (int i = 0; i < RefState.N; ++i)
      answer += 1.0/(pow(lambdaoc - RefState.lambdaoc[i], 2.0) + delta*delta);
      answer *= delta/(PI * RefState.L);

    return(answer);
  }

  // Better implementation: making use of rediscretized TBA state.
  LiebLin_Bethe_State Canonical_Saddle_Point_State (DP c_int, DP L, int N, DP kBT)
  {
    // This function returns the discretized state minimizing the canonical free energy
    // F = E - T S.

    // This is obtained by rediscretizing the solution coming from TBA.

    // ASSUMPTIONS: 
    // Periodic boundary conditions (the state which is output is forced to be symmetric Ix2 == -Ix2).

    // For zero temperature, return the ground state:
    if (fabs(kBT) < 1.0e-4) return(LiebLin_Bethe_State(c_int, L, N));

    // Otherwise, return the discretized TBA saddle-point state:
    LiebLin_TBA_Solution TBAsol = LiebLin_TBA_Solution_fixed_nbar (c_int, N/L, kBT, 1.0e-4, JSC::max(N/10, 10));

    LiebLin_Bethe_State spstate = Discretized_LiebLin_Bethe_State (c_int, L, N, TBAsol.rho);

    // Explicitly symmetrize:
    for (int i = 0; i < N/2; ++i) spstate.Ix2[i] = -spstate.Ix2[N-1-i];
    spstate.Compute_All(false);

    return(spstate);
  }

  LiebLin_Bethe_State Add_Particle_at_Center (const LiebLin_Bethe_State& RefState)
  {
    LiebLin_Bethe_State ReturnState (RefState.c_int, RefState.L, RefState.N + 1);

    // Add a quantum number in middle (explicitly: to right of index N/2) 
    // and shift quantum numbers by half-integer away from added one:
    ReturnState.Ix2[RefState.N/2] = RefState.Ix2[RefState.N/2] - 1;
    for (int i = 0; i < RefState.N+1; ++i)
      ReturnState.Ix2[i + (i >= RefState.N/2)] = RefState.Ix2[i] - 1 + 2*(i >= RefState.N/2);

    return(ReturnState);
  }

  LiebLin_Bethe_State Remove_Particle_at_Center (const LiebLin_Bethe_State& RefState)
  {
    LiebLin_Bethe_State ReturnState (RefState.c_int, RefState.L, RefState.N - 1);

    // Remove midmost and shift quantum numbers by half-integer towards removed one:
    for (int i = 0; i < RefState.N-1; ++i)
      ReturnState.Ix2[i] = RefState.Ix2[i + (i >= RefState.N/2)] + 1 - 2*(i >= RefState.N/2);

    return(ReturnState);
  }

} // namespace JSC
