/**********************************************************

This software is part of J.-S. Caux's ABACUS library.

Copyright (c).

-----------------------------------------------------------

File:  src/LIEBLIN/ln_g2_ME.cc

Purpose:  provides the matrix element of the second density moment g2(x = 0)

NOTE:  based on Lorenzo Piroli's expression of the matrix element


***********************************************************/

#include "JSC.h"

using namespace std;
using namespace JSC;

namespace JSC {
/*
  complex<DP> ln_g2_ME_old_jacopo (LiebLin_Bethe_State& mu, LiebLin_Bethe_State& lambda)
  {	
    
    if (mu.Ix2 == lambda.Ix2) return(-200.);
    
    DP c_int = mu.c_int;
    
    SQMat_CX G(lambda.N);
    SQMat_CX GpB(lambda.N);
    SQMat_CX B(lambda.N);
    
    complex<DP> log_themp;
    complex<DP> lnprefactor = 0.  ;
    
    lnprefactor += 2.*log(c_int) ;
    
    for(int j=0; j < mu.N; ++j){
      for(int k=0; k < j; ++k){
	lnprefactor -= log((mu.lambdaoc[j] - mu.lambdaoc[k]));
	lnprefactor -= log((lambda.lambdaoc[j] - lambda.lambdaoc[k]));
      }
    }
    
    
    for(int j=0; j < mu.N; ++j){
      for(int k=0; k <mu.N; ++k){
	lnprefactor += log(mu.lambdaoc[j] - mu.lambdaoc[k] + II);
      }
    }
    
    Vect_CX VVP (0.0, mu.N);
    Vect_CX VVM (0.0, mu.N);
    
    //compute the vectors
    for(int j=0; j < mu.N; ++j) {
      log_themp = 0.;
      for(int k=0; k < mu.N; ++k) log_themp += log(mu.lambdaoc[j] - lambda.lambdaoc[k] + II) - log(mu.lambdaoc[j] - mu.lambdaoc[k] + II);
      VVP[j] = exp(log_themp);
    }
    
    for(int j=0; j < mu.N; ++j) {
      log_themp = 0.;
      for(int k=0; k < mu.N; ++k) log_themp += log(mu.lambdaoc[j] - lambda.lambdaoc[k] - II) - log(mu.lambdaoc[j] - mu.lambdaoc[k] - II);
      VVM[j] = exp(log_themp);
    }
    
    //compute the sum of determinants
    complex<DP> sum_n =0.;
    for(int n=0; n < mu.N; ++n) {
        // compute the matrices
      for (int a = 0; a < mu.N; ++a){
	for (int b = 0; b < mu.N; ++b){
	  if(b == n){
	    G[a][b] = 2.*II* lambda.lambdaoc[a];
	  }
	  else{
	    G[a][b] = II*1./(lambda.lambdaoc[a] - mu.lambdaoc[b])*( VVP[b]*1.0/(mu.lambdaoc[b] - lambda.lambdaoc[a] + II)  +
								VVM[b]*1.0/(mu.lambdaoc[b] - lambda.lambdaoc[a] - II)  );
	  }
	}
      }
        
      for (int a = 0; a < mu.N; ++a){
	for (int b = 0; b < mu.N; ++b){
	  if(b == n){
	    B[a][b] = 0.;
	  }
	  else{ 
	    B[a][b] = II*1.0/(mu.lambdaoc[b] - mu.lambdaoc[n] - II) ;
	  }
	}
      } 
        
      for (int a = 0; a < mu.N; ++a){ 
	for (int b = 0; b < mu.N; ++b){
	  GpB[a][b] = G[a][b] + B[a][b];
	}
      } 
      //finally add
      sum_n +=   exp(	lndet_LU_CX(GpB) ) -  exp(lndet_LU_CX(G) ) ;
    }
    
    return(log(sum_n ) + lnprefactor  - 0.5 * (mu.lnnorm + lambda.lnnorm));
  }
*/
    complex<DP> Fn_V_g2 (int j, LiebLin_Bethe_State& lstate, LiebLin_Bethe_State& rstate)
    {
        //complex<DP> result_num = 1.0;
        //complex<DP> result_den = 1.0;
        complex<DP> result = 1.0;
        
        for (int m = 0; m < lstate.N; ++m) {
            //result_num *= (lstate.lambdaoc[m] - rstate.lambdaoc[j] + II);
            //result_den *= (rstate.lambdaoc[m] - rstate.lambdaoc[j] + II);
            result *= (lstate.lambdaoc[m] - rstate.lambdaoc[j] + II)/(rstate.lambdaoc[m] - rstate.lambdaoc[j] + II);
        }
        
        //return(result_num/result_den);
        return(result);
    }

    
    complex<DP> ln_g2_ME (LiebLin_Bethe_State& lstate, LiebLin_Bethe_State& rstate)
    {
        // Computes the log of the density operator \rho(x = 0) matrix element between lstate and rstate.
        
        // If we have lstate == rstate, density matrix element = 0:
        
        
        if (lstate.Ix2 == rstate.Ix2) return(-200.);
        
      
        
        
        SQMat_DP one_plus_U (0.0, lstate.N);
        
        Vect_CX Vplus (0.0, lstate.N);  // contains V^+_j;  V^-_j is the conjugate
        //Vect_DP Fn_Prod (0.0, lstate.N);  // product_{m\neq j} (\mu_m - \lambdaoc_j)/(\lambdaoc_m - \lambdaoc_j)
        // From ABACUS++G_3 onwards: use logs to stabilize numerical values at small c, at the cost of execution speed.
        Vect_CX ln_Fn_Prod (0.0, lstate.N);  // product_{m\neq j} (\mu_m - \lambdaoc_j)/(\lambdaoc_m - \lambdaoc_j)
        Vect_DP rKern (0.0, lstate.N);   // K(lambdaoc_j - lambdaoc_(p == arbitrary))
        
        //int p = 0;
        int p = rstate.N/2-1; // choice doesn't matter, see 1990_Slavnov_TMP_82 after (3.8). Choose rapidity around the middle.
        
        
        DP c_int = rstate.c_int;
        
        DP Eout = log(pow(lstate.E - rstate.E,2.)) - (2)*log(c_int)  - log(2*lstate.N);
        
        for (int a = 0; a < lstate.N; ++a) {
            Vplus[a] = Fn_V_g2 (a, lstate, rstate);
            //Fn_Prod[a] = 1.0;
            ln_Fn_Prod[a] = 0.0;;
            for (int m = 0; m < lstate.N; ++m)
                //if (m != a) Fn_Prod[a] *= (lstate.lambdaoc[m] - rstate.lambdaoc[a])/(rstate.lambdaoc[m] - rstate.lambdaoc[a]);
                if (m != a) ln_Fn_Prod[a] += log(complex<DP>(lstate.lambdaoc[m] - rstate.lambdaoc[a])/(rstate.lambdaoc[m] - rstate.lambdaoc[a]));
            rKern[a] = rstate.Kernel (a, p);
        }
        
        for (int a = 0; a < lstate.N; ++a)
            for (int b = 0; b < lstate.N; ++b)
                one_plus_U[a][b] = (a == b ? 1.0 : 0.0) + 0.5 * 1./imag(Vplus[a])
                //* Fn_Prod[a] * (rstate.Kernel(a,b) - rKern[b]);
                * ((lstate.lambdaoc[a] - rstate.lambdaoc[a])*real(exp(ln_Fn_Prod[a])) * (rstate.Kernel(a,b) - rKern[b]) + rKern[b]);
        
        complex<DP> ln_ddalpha_sigma = lndet_LU_dstry(one_plus_U);
        
        complex<DP> ln_prod_V = 0.0;
        for (int a = 0; a < lstate.N; ++a) ln_prod_V += log(2.0 * II * imag(Vplus[a]));
        
        complex<DP> ln_prod_2 = 0.0;
        for (int a = 0; a < lstate.N; ++a)
            for (int b = 0; b < lstate.N; ++b)
                ln_prod_2 += log((rstate.lambdaoc[a] - rstate.lambdaoc[b] + II)/(lstate.lambdaoc[a] - rstate.lambdaoc[b]));
        
        //cout << "ln_Fn_Prod = " << ln_Fn_Prod << endl;
        //cout << endl << ln_ddalpha_sigma << "\t" << ln_prod_V << "\t" << ln_prod_2 << "\t" << - log(2.0 * II * imag(Vplus[p])) << "\t" << - 0.5 * (lstate.lnnorm + rstate.lnnorm) << "\t" << ln_ddalpha_sigma + ln_prod_V + ln_prod_2 - log(2.0 * II * imag(Vplus[p])) - 0.5 * (lstate.lnnorm + rstate.lnnorm) << endl;
        
        ln_ddalpha_sigma += ln_prod_V + ln_prod_2 - log(2.0 * II * imag(Vplus[p]));
        
        return (log(II) + Eout + ln_ddalpha_sigma - 0.5 * (lstate.lnnorm + rstate.lnnorm));
    }

    
    
    
    
} // namespace JSC
