/**********************************************************

This software is part of J.-S. Caux's ABACUS library.

Copyright (c).

-----------------------------------------------------------

File:  Scan_Thread_List.cc

Purpose:  defines all functions for Scan_Thread_List class.
          

***********************************************************/

#include "JSC.h"

using namespace std;
using namespace JSC;

namespace JSC {

  Scan_Thread_List::Scan_Thread_List()
  {
    dim = 2;
    nthreads_tot = 0;
    nthreads_done = 0;
    //omega = Vect<float>(1.0e+6, dim);
    //iK = Vect_INT(0, dim);
    //abs_data_value = Vect_DP (0.0, Nentries);
    abs_data_value = Vect<float> (0.0, dim);
    label = Vect<string> (dim);
    type = Vect<int> (dim);
    isdone = Vect<bool> (false, dim);
  }

  Scan_Thread_List::Scan_Thread_List (int Nentries)
  {
    dim = Nentries;
    nthreads_tot = 0;
    nthreads_done = 0;
    ////omega = Vect_DP(1.0e+6, Nentries);
    //omega = Vect<float>(1.0e+6, Nentries);
    //iK = Vect_INT(0, Nentries);
    //abs_data_value = Vect_DP (0.0, Nentries);
    abs_data_value = Vect<float> (0.0, Nentries);
    label = Vect<string> (Nentries);
    type = Vect<int> (Nentries);
    isdone = Vect<bool> (false, Nentries);
  }

  bool Scan_Thread_List::Increase_Size (int nr_to_add)  // resizes the vectors to accommodate up to nr_to_add additional entries
  {
    dim += nr_to_add;
    // Nentries stays the same

    //cout << "Called Increase_Size with nr_to_add = " << nr_to_add << endl;
    try {
      //omega.Increase_Size (nr_to_add);
      //iK.Increase_Size (nr_to_add);
      abs_data_value.Increase_Size (nr_to_add);
      //cout << "Increasing size of label" << endl;
      label.Increase_Size (nr_to_add, "");
      //cout << "Done increasing size of label" << endl;
      type.Increase_Size (nr_to_add);
      isdone.Increase_Size (nr_to_add, false);
    }

    catch (bad_alloc) {
      cout << "dim " << dim << "\tnr_to_add " << nr_to_add << endl;
      JSCerror("Memory allocation failed in Scan_Thread_List::Increase_Size.");
    }

    return(true);
  }

  //void Scan_Thread_List::Include_Thread (DP omega_ref, int iK_ref, DP abs_data_value_ref, string label_ref, int type_ref) 
  void Scan_Thread_List::Include_Thread (DP data_value_ref, string label_ref, int type_ref) 
  {
    if (nthreads_tot > dim - 10) { // Resize the Scan_Thread_List, by doubling its size
      //cout << "Resizing threads list" << endl;
      (*this).Increase_Size (dim);
      //cout << "Done resizing" << endl;
    }

    //cout << "Including thread for label " << label_ref << " and type " << type_ref << endl;

    //omega[nthreads] = omega_ref;
    //iK[nthreads] = iK_ref;
    abs_data_value[nthreads_tot] = abs(data_value_ref);
    label[nthreads_tot] = label_ref;
    type[nthreads_tot] = type_ref;
    isdone[nthreads_tot] = false;

    nthreads_tot++;

    if (nthreads_tot >= dim) {
      (*this).Save("Threads_stopped.thr");
      JSCerror("nthreads_tot >= dim in Scan_Thread_List.");
    }

  }

  void Scan_Thread_List::Order_in_abs_data_value ()
  {
    // Reorders all entries in decreasing order in abs_data_value

    if (nthreads_tot > 1) {

      Vect_INT index(dim);
      for (int i = 0; i < nthreads_tot; ++i) index[i] = i;

      abs_data_value.QuickSort(index, 0, nthreads_tot - 1);

      ////Vect_DP omega_ordered(nthreads);
      //Vect<float> omega_ordered(nthreads);
      //Vect_INT iK_ordered(nthreads);
      //Vect_DP abs_data_value_ordered(nthreads);
      Vect<float> abs_data_value_ordered(nthreads_tot);
      Vect<string> label_ordered(nthreads_tot);
      Vect<int> type_ordered(nthreads_tot);
      Vect<bool> isdone_ordered(nthreads_tot);

      // Put data in proper order
      for (int i = 0; i < nthreads_tot; ++i) {
	//omega_ordered[i] = omega[index[nthreads - 1 - i] ];
	//iK_ordered[i] = iK[index[nthreads - 1 - i] ];
	abs_data_value_ordered[i] = abs_data_value[nthreads_tot - 1 - i];
	label_ordered[i] = label[index[nthreads_tot - 1 - i] ];
	type_ordered[i] = type[index[nthreads_tot - 1 - i] ];
	//sector_lowest_raisable_ordered[i] = sector_lowest_raisable[index[nthreads - 1 - i] ];
	isdone_ordered[i] = isdone[index[nthreads_tot - 1 - i] ];
      }
      
      // Put back in *this object:
      for (int i = 0; i < nthreads_tot; ++i) {
	//omega[i] = omega_ordered[i];
	//iK[i] = iK_ordered[i];
	abs_data_value[i] = abs_data_value_ordered[i];
	label[i] = label_ordered[i];
	type[i] = type_ordered[i];
	isdone[i] = isdone_ordered[i];
      } // The rest are all simply 0.

      // Done !
    }
  }
  /*
  void Scan_Thread_List::Order_in_omega ()
  {
    // Reorders all entries in increasing order in omega

    if (nthreads > 1) {

      Vect_INT index(0, dim);
      for (int i = 0; i < nthreads; ++i) index[i] = i;

      omega.QuickSort(index, 0, nthreads - 1);
      
      //Vect_DP omega_ordered(nthreads);
      Vect<float> omega_ordered(nthreads);
      Vect_INT iK_ordered(nthreads);
      //Vect_DP abs_data_value_ordered(nthreads);
      Vect<float> abs_data_value_ordered(nthreads);
      Vect<string> label_ordered(nthreads);
      Vect<int> type_ordered(nthreads);

      // Put data in proper order
      for (int i = 0; i < nthreads; ++i) {
	omega_ordered[i] = omega[i];
	iK_ordered[i] = iK[index[i] ];
	abs_data_value_ordered[i] = abs_data_value[index[i] ];
	label_ordered[i] = label[index[i] ];
	type_ordered[i] = type[index[i] ];
      }
      
      // Put back in *this object:
      for (int i = 0; i < nthreads; ++i) {
	omega[i] = omega_ordered[i];
	iK[i] = iK_ordered[i];
	abs_data_value[i] = abs_data_value_ordered[i];
	label[i] = label_ordered[i];
	type[i] = type_ordered[i];
      } // The rest are all simply 0.

      // Done !
    }
  }
  */
  /*
  DP Scan_Thread_List::Highest_abs_data_value_ordered (DP omega_MIN, DP omega_MAX) // returns highest abs_data_value in window omega_MIN to omega_MAX
  {
    // Assume that the threads have been Order_in_abs_data_value
    for (int i = 0; i < nthreads; ++i) 
      if (omega[i] > omega_MIN && omega[i] < omega_MAX) return(abs_data_value[i]);

    return(0.0);
  }
  */
  //DP Scan_Thread_List::Highest_abs_data_value (DP omega_MIN, DP omega_MAX) // returns highest abs_data_value in window omega_MIN to omega_MAX
  DP Scan_Thread_List::Highest_abs_data_value () // returns highest abs_data_value
  {
    // Works even if list is not ordered
    DP maxdatavalue = 0.0;
    for (int i = 0; i < nthreads_tot; ++i) 
      //if (omega[i] > omega_MIN && omega[i] < omega_MAX && abs_data_value[i] > maxdatavalue) 
      if (!isdone[i] && (abs_data_value[i] > maxdatavalue))
	maxdatavalue = abs_data_value[i];

    return(maxdatavalue);
  }

  DP Scan_Thread_List::kth_highest_abs_data_value (int k) // returns the k-th highest abs_data_value in window omega_MIN to omega_MAX
  {
    // Works even if list is not ordered
    if (k < 1) JSCerror("Give k > 1 in Scan_Thread_List::kth_Highest_abs_data_value.");
    else if (k == 1 || k >= nthreads_tot) // Threads list not long enough, return the top one
      //return ((*this).Highest_abs_data_value (-1.0e+10, 1.0e+10));
      return ((*this).Highest_abs_data_value ());

    Vect<DP> topk (0.0, k);

    for (int i = 0; i < nthreads_tot; ++i) 
      if (!isdone[i] && (abs_data_value[i] > topk[0])) {
	topk[0] = abs_data_value[i];
	topk.QuickSort(0, k-1);
      }

    return(topk[0]);
  }

  bool Scan_Thread_List::Exists_data_value_greater_than (DP value)
  {
    for (int i = 0; i < nthreads_tot; ++i)
      if (abs_data_value[i] > value) return(true);
    return(false);
  }

  /*
  DP Scan_Thread_List::Lowest_omega () // returns lowest omega
  {
    // Assume that the threads have been Order_in_omega
    return(omega[0]);
  }
  */
  void Scan_Thread_List::Merge (const Scan_Thread_List& reflist)
  {
    if (nthreads_tot + reflist.nthreads_tot >= dim) // JSCerror("Scan_Threads_List:  too big to merge.");
      //(*this).Increase_Size (reflist.nthreads);
      (*this).Increase_Size (nthreads_tot/10 + reflist.nthreads_tot);

    for (int i = 0; i < reflist.nthreads_tot; ++i) {
      //omega[nthreads] = reflist.omega[i];
      //iK[nthreads] = reflist.iK[i];
      abs_data_value[nthreads_tot] = reflist.abs_data_value[i];
      label[nthreads_tot] = reflist.label[i];
      type[nthreads_tot] = reflist.type[i];
      isdone[nthreads_tot] = reflist.isdone[i];
      nthreads_tot++;
    }
  }

  /* // DEACTIVATED IN ABACUS++G_2
  void Scan_Thread_List::Remove_Threads (int ithread_down, int ithread_up)
  {
    if (ithread_down < 0 || ithread_up > nthreads) { 
      JSCerror("Trying to remove inexistent entries in Remove_Threads.");
    }

    for (int i = 0; i < nthreads - ithread_up - 1; ++i) {
      //omega[ithread_down + i] = omega[ithread_up + i + 1];
      //iK[ithread_down + i] = iK[ithread_up + 1 + i];
      abs_data_value[ithread_down + i] = abs_data_value[ithread_up + i + 1];
      label[ithread_down + i] = label[ithread_up + i + 1];
      type[ithread_down + i] = type[ithread_up + i + 1];
    }
    // Zero the other entries:  
    for (int i = nthreads - ithread_up - 1; i < nthreads; ++i) {
      //omega[i] = 1.0e+6;
      //iK[i] = 0;
      abs_data_value[i] = 0.0;//-1.0e-30;  // give a small negative value, so these are put at the very bottom of the list
      label[i] = "";
      type[i] = 0;
    }

    nthreads -= ithread_up - ithread_down + 1;  // removed that many entries.
  }
  */
  /* // DEACTIVATED IN ABACUS++G_2
  int Scan_Thread_List::Remove_Threads (const Vect<bool>& threads_done)
  {
    // If threads_done[ithread] == true, remove from list

    if (threads_done.size() != nthreads) JSCerror("Wrong size boolean vector in Scan_Thread_List::Remove_Threads.");

    int nr_removed = 0;

    for (int i = 0; i < nthreads; ++i) {
      if (!threads_done[i]) {
	//omega[i - nr_removed] = omega[i];
	//iK[i - nr_removed] = iK[i];
	abs_data_value[i - nr_removed] = abs_data_value[i];
	label[i - nr_removed] = label[i];
	type[i - nr_removed] = type[i];
	
      }
      else nr_removed++;
    }
    // Zero other entries:
    for (int i = nthreads - nr_removed; i < nthreads; ++i) {
      //omega[i] = 1.0e+6;
      //iK[i] = 0;
	abs_data_value[i] = 0.0;
	label[i] = "";
	type[i] = 0;
    }

    nthreads -= nr_removed;

    return(nr_removed);
  }
  */

  void Scan_Thread_List::Remove_Done_Threads ()
  {
    // If isdone[ithread] == true, remove from list

    int nr_removed = 0;

    for (int i = 0; i < nthreads_tot; ++i) {
      if (!isdone[i]) {
	//omega[i - nr_removed] = omega[i];
	//iK[i - nr_removed] = iK[i];
	abs_data_value[i - nr_removed] = abs_data_value[i];
	label[i - nr_removed] = label[i];
	type[i - nr_removed] = type[i];
	isdone[i - nr_removed] = isdone[i];
      }
      else nr_removed++;
    }
    // Zero other entries:
    for (int i = nthreads_tot - nr_removed; i < nthreads_tot; ++i) {
      //omega[i] = 1.0e+6;
      //iK[i] = 0;
	abs_data_value[i] = 0.0;
	label[i] = "";
	type[i] = 0;
	isdone[i] = false;
    }

    nthreads_tot -= nr_removed;
    if (nthreads_done != nr_removed) {
      cout << nthreads_done << "\t" << nr_removed << endl;
      JSCerror("Miscount in removing threads during Remove_Done_Threads.");
    }
    nthreads_done = 0;

    return;
  }

  void Scan_Thread_List::Clear ()
  {
    // dim is preserved

    nthreads_tot = 0;
    nthreads_done = 0;
    ////omega = Vect_DP(0.0, dim);
    //omega = Vect<float> (1.0e+6, dim);
    //iK = Vect_INT(0, dim);
    //abs_data_value = Vect_DP (0.0, dim);
    abs_data_value = Vect<float> (0.0, dim);
    label = Vect<string> (dim);
    type = Vect<int> (dim);
    isdone = Vect<bool> (false, dim);
  }

  void Scan_Thread_List::Save(const char* outfile_Cstr)
  {
    // We save only the undone threads, so after cleanup.
    (*this).Remove_Done_Threads();

    ofstream outfile;
    
    outfile.open(outfile_Cstr);
    if (outfile.fail()) JSCerror("Could not open outfile... "); 
    outfile.precision(3);
    
    if (nthreads_tot > 0) {
      //outfile << omega[0] << "\t" << iK[0] << "\t" << abs_data_value[0] << "\t" << label[0] << "\t" << type[0];
      outfile << abs_data_value[0] << "\t" << label[0] << "\t" << type[0];
      for (int i = 1; i < nthreads_tot; ++i) {
	//outfile << endl << omega[i] << "\t" << iK[i] << "\t" << abs_data_value[i] << "\t" << label[i] << "\t" << type[i];
	outfile << endl << abs_data_value[i] << "\t" << label[i] << "\t" << type[i];
      }
    }

    outfile.close();

    return;
  }

  void Scan_Thread_List::Load (const char* thrfile_Cstr)
  {
    /*
    // Check that the data fits within limits:
    struct stat statbuf;

    stat (thrfile_Cstr, &statbuf);
    int filesize = statbuf.st_size;

    // Determine the number of entries approximately
    // Don't overestimate entry_size: pretend label takes no space...
    int entry_size = 2* sizeof(float) + sizeof(int);

    int estimate_nr_entries = filesize/entry_size;
    */

    // Get number of lines in the file:
    int nrlines = 0;
    ifstream infile1(thrfile_Cstr);
    if(infile1.fail()) {
      cout << "Could not open file " << thrfile_Cstr << " in Scan_Thread_List::Load." << endl;
      JSCerror("Terminating.");
    }
    string dummy;
    while (getline(infile1, dummy))  ++nrlines;
    infile1.close();

    int estimate_nr_entries = JSC::max(10, (11*nrlines)/10);

    if (estimate_nr_entries > dim) {
      (*this).Increase_Size (estimate_nr_entries - dim/2);  // give dim/2 safety
    }

    ifstream infile;
    infile.open(thrfile_Cstr);
    if(infile.fail()) {
      cout << "Could not open file " << thrfile_Cstr << " in Scan_Thread_List::Load." << endl;
      JSCerror("Terminating.");
    }

    (*this).Clear();

    DP abs_data_value_DP;
    while (infile.peek() != EOF) {
      //infile >> omega[nthreads];
      //infile >> iK[nthreads];
      //infile >> abs_data_value[nthreads];  // nasty bug:  reading into float bugs when value less or more than float limits
      infile >> abs_data_value_DP;   // cure:  first read into double, then cast to float:
      abs_data_value[nthreads_tot] = float(abs_data_value_DP);
      infile >> label[nthreads_tot];
      infile >> type[nthreads_tot];
      isdone[nthreads_tot] = false;
      nthreads_tot++;
      if (nthreads_tot >= dim) {
	cout << "file " << thrfile_Cstr << endl << "estimate_nr_entries = " << estimate_nr_entries << "\tnrlines = " << nrlines << endl << "nthreads_tot = " << nthreads_tot << "\tdim = " << dim << endl;
	JSCerror("Too many threads in input file.");
      }
    }

    infile.close();

    return;
  }

} // namespace JSC
