/**********************************************************

This software is part of J.-S. Caux's C++ library.

Copyright (c) 2006.

-----------------------------------------------------------

File:  SF_4p_client.cc

Purpose:  4 spinon contrib to S(k, omega) for XXX_h0, parallel code using MPICH.
          Client part.

Last modified:  13/10/06

***********************************************************/

#include "mpi.h"
#include "JSC.h"

namespace JSC {

  void SF_4p_opt_client (MPI_Comm comm, I_table Itable)
  {  
    int server = 0;
    int rank; 

    MPI_Status status;

    int client_request_size = 7;
    DP client_request[client_request_size]; 
    // this is:
    // ik
    // iomega
    // k
    // omega
    // req_prec
    // Npts_K
    // Npts_W
    int client_result_size = 8; 
    DP client_result[client_result_size]; // as above, + SF_4p

    MPI_Comm_rank (comm, &rank);

    MPI_Recv (client_request, client_request_size, MPI_DOUBLE, server, MPI_ANY_TAG, comm, &status);
    //cout << "Client of rank " << rank << ":  received request ";
    //for (int i = 0; i < client_request_size; ++i) cout << client_request[i] << " ";
    //cout << "with tag " << status.MPI_TAG << endl;


    while (status.MPI_TAG) {

      // Result:  SF_4p_opt (k, omega)
      client_result[0] = client_request[0];
      client_result[1] = client_request[1];
      client_result[2] = client_request[2];
      client_result[3] = client_request[3];
      client_result[4] = client_request[4];
      client_result[5] = client_request[5];
      client_result[6] = client_request[6];
      client_result[7] = SF_4p_opt (client_request[2], client_request[3], client_request[4], 
				    int(client_request[5]), int(client_request[6]), Itable);
      MPI_Send (client_result, client_result_size, MPI_DOUBLE, server, status.MPI_TAG, comm);
      //cout << "Client of rank " << rank << " sending complete" << endl;
      
      // Wait for subsequent request from server
      MPI_Recv (client_request, client_request_size, MPI_DOUBLE, server, MPI_ANY_TAG, comm, &status);
      //cout << "Client of rank " << rank << ":  received request ";
      //for (int i = 0; i < client_request_size; ++i) cout << client_request[i] << " ";
      //cout << "with tag " << status.MPI_TAG << endl;

    }

    return;
  }
  /*
  void SF_4p_kwKW_alpha_client (MPI_Comm comm, DP req_prec, int max_rec, I_table Itable)
  {  
    int server = 0;
    int rank; 

    MPI_Status status;

    int client_request_size = 2;
    DP client_request[client_request_size]; // this is k, alpha
    int client_result_size = 3;
    DP client_result[client_result_size]; // this is k, alpha, SF_4p_kwKW_alpha

    MPI_Comm_rank (comm, &rank);

    MPI_Recv (client_request, client_request_size, MPI_DOUBLE, server, MPI_ANY_TAG, comm, &status);
    //cout << "Client of rank " << rank << ":  received request ";
    //for (int i = 0; i < client_request_size; ++i) cout << client_request[i] << " ";
    //cout << "with tag " << status.MPI_TAG << endl;

    Vect_DP args_to_SF_4p_kwKW_alpha(4);
    args_to_SF_4p_kwKW_alpha[2] = req_prec;
    args_to_SF_4p_kwKW_alpha[3] = DP(max_rec);

    while (status.MPI_TAG) {

      args_to_SF_4p_kwKW_alpha[0] = client_request[0];
      args_to_SF_4p_kwKW_alpha[1] = client_request[1];

      // Result:  SF_4p (k, omega)
      client_result[0] = client_request[0];
      client_result[1] = client_request[1];
      client_result[2] = SF_4p_kwKW_alpha (args_to_SF_4p_kwKW_alpha, Itable);

      MPI_Send (client_result, client_result_size, MPI_DOUBLE, server, status.MPI_TAG, comm);
      //cout << "Client of rank " << rank << " sending complete" << endl;
      
      // Wait for subsequent request from server
      MPI_Recv (client_request, client_request_size, MPI_DOUBLE, server, MPI_ANY_TAG, comm, &status);
      //cout << "Client of rank " << rank << ":  received request ";
      //for (int i = 0; i < client_request_size; ++i) cout << client_request[i] << " ";
      //cout << "with tag " << status.MPI_TAG << endl;

    }

    return;
  }

  void SF_4p_kwKW_alpha_opt_client (MPI_Comm comm, DP req_prec, int Npts_K, int Npts_W, I_table Itable)
  {  
    int server = 0;
    int rank; 

    MPI_Status status;

    int client_request_size = 2;
    DP client_request[client_request_size]; // this is k, alpha
    int client_result_size = 3;
    DP client_result[client_result_size]; // this is k, alpha, SF_4p_kwKW_alpha_opt

    MPI_Comm_rank (comm, &rank);

    MPI_Recv (client_request, client_request_size, MPI_DOUBLE, server, MPI_ANY_TAG, comm, &status);
    //cout << "Client of rank " << rank << ":  received request ";
    //for (int i = 0; i < client_request_size; ++i) cout << client_request[i] << " ";
    //cout << "with tag " << status.MPI_TAG << endl;

    Vect_DP args_to_SF_4p_kwKW_alpha_opt(5);
    args_to_SF_4p_kwKW_alpha_opt[2] = req_prec;
    args_to_SF_4p_kwKW_alpha_opt[3] = DP(Npts_K);
    args_to_SF_4p_kwKW_alpha_opt[4] = DP(Npts_W);

    while (status.MPI_TAG) {

      args_to_SF_4p_kwKW_alpha_opt[0] = client_request[0];
      args_to_SF_4p_kwKW_alpha_opt[1] = client_request[1];

      // Result:  SF_4p (k, omega)
      client_result[0] = client_request[0];
      client_result[1] = client_request[1];
      client_result[2] = SF_4p_kwKW_alpha_opt (args_to_SF_4p_kwKW_alpha_opt, Itable);

      MPI_Send (client_result, client_result_size, MPI_DOUBLE, server, status.MPI_TAG, comm);
      //cout << "Client of rank " << rank << " sending complete" << endl;
      
      // Wait for subsequent request from server
      MPI_Recv (client_request, client_request_size, MPI_DOUBLE, server, MPI_ANY_TAG, comm, &status);
      //cout << "Client of rank " << rank << ":  received request ";
      //for (int i = 0; i < client_request_size; ++i) cout << client_request[i] << " ";
      //cout << "with tag " << status.MPI_TAG << endl;

    }

    return;
  }
  */
} // namespace JSC
